
// fx - A compositing library representation.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#include <exception>
#include <string>

#if _MSC_VER >= 1400
#include <hash_map>
#else
#include <map>
#endif

#include <openpluginlib/pl/property.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#include <openeffectslib/fx/n_operator_gpu_kernel_importer.hpp>
#include <openeffectslib/fx/n_operator_gpu_kernel.hpp>

namespace fs  = boost::filesystem;
namespace opl = olib::openpluginlib;

namespace olib { namespace openeffectslib { namespace fx {

namespace
{
	template<class T>
	struct value_
	{
		void operator( )( T* p, typename T::value_type value )
		{ p->get( ) = value; }
	};

	class p_value
	{
		typedef opl::wstring key_type;
		typedef boost::any	 func_type;
		
	public:
#if _MSC_VER >= 1400
		typedef stdext::hash_multimap<key_type, func_type>	container;
#else
		typedef std::multimap<key_type, func_type>			container;
#endif

		typedef container::const_iterator	const_iterator;
		typedef container::value_type		value_type;

	public:
		explicit p_value( )
		{
			func_.insert( value_type( opl::v_bool_typename, value_<opl::v_bool>( ) ) );
		}
		
	private:
		container func_;
	};
	
	class p_factory
	{
	public:
		typedef opl::wstring	key_type;
		typedef boost::any		param_type;
		
#if _MSC_VER >= 1400
		typedef stdext::hash_multimap<key_type, param_type> container;
#else
		typedef std::multimap<key_type, param_type>			container;
#endif

		typedef container::const_iterator	const_iterator;
		typedef container::value_type		value_type;

	public:
		explicit p_factory( )
		{
			reg_.insert( value_type( opl::v_bool_typename,			opl::v_bool( ) ) );
			reg_.insert( value_type( opl::v_color_typename,			opl::v_color( ) ) );
			reg_.insert( value_type( opl::v_color_rgba_typename,	opl::v_color_rgba( ) ) );
			reg_.insert( value_type( opl::v_double_typename,		opl::v_double( ) ) );
			reg_.insert( value_type( opl::v_float_typename,			opl::v_float( ) ) );
			reg_.insert( value_type( opl::v_int32_typename,			opl::v_int32( ) ) );
			reg_.insert( value_type( opl::v_uint32_typename,		opl::v_uint32( ) ) );
			reg_.insert( value_type( fx::v_node_typename,			fx::v_node_ptr( ) ) );
		}
		
		value_type value( const key_type& key ) const
		{
			const_iterator I = reg_.find( key  );
			if( I != reg_.end( ) )
				return *I;
				
			throw std::exception( );
		}
		
	private:
		container reg_;
	};

	fx::n_operator_gpu::creator_t creator( const fs::path& path )
	{
		typedef n_operator_gpu::container			container;
		typedef n_operator_gpu::value_type			value_type;
		typedef n_operator_gpu::param_container		param_container;
		typedef n_operator_gpu::param_value_type	param_value_type;

		fx::n_operator_gpu::creator_t creator_;
		//p_factory p_creator_;

		n_operator_gpu_kernel_importer importer;
		if( !importer( path ) ) return creator_;

		typedef gpu_kernel_item::const_iterator const_iterator;
		for( const_iterator I = importer.kernel.param.begin( ); I != importer.kernel.param.end( ); ++I )
		{
			if( I->type == opl::v_bool_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_bool( ) ) );
			}
			else if( I->type == opl::v_color_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_color( ) ) );
			}
			else if( I->type == opl::v_color_rgba_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_color_rgba( ) ) );
			}
			else if( I->type == opl::v_double_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_double( ) ) );
			}
			else if( I->type == opl::v_float_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_float( ) ) );
			}
			else if( I->type == opl::v_int32_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_int32( ) ) );
			}
			else if( I->type == opl::v_uint32_typename )
			{
				creator_.first.insert( value_type( I->name,  opl::v_uint32( ) ) );
			}
			else if( I->type == fx::v_in_color_typename )
			{
				creator_.first.insert( value_type( I->name,  fx::v_node_ptr( ) ) );
			}
			else if( I->type == fx::v_out_color_typename )
			{
				creator_.first.insert( value_type( I->name,  fx::v_node_ptr( ) ) );
			}
			
			creator_.second.insert( param_value_type( I->name, I->type ) );
		}

		if( importer.kernel.GLSL_valid )
		{
			creator_.first.insert( value_type(			L"GLSL_url",  opl::v_url( fs::path( opl::to_string( importer.kernel.GLSL_language.url ).c_str( ), fs::native ) ) ) );
			creator_.second.insert( param_value_type(	L"GLSL_url",  opl::v_url_typename ) );
		}
		
		if( importer.kernel.Cg_valid )
		{
			creator_.first.insert( value_type(			L"Cg_url",	  opl::v_url( fs::path( opl::to_string( importer.kernel.Cg_language.url ).c_str( ), fs::native ) ) ) );
			creator_.second.insert( param_value_type(	L"Cg_url",    opl::v_url_typename ) );
		}

		return creator_;
	}
}

namespace
{
	const openpluginlib::shader_string GLSL_vert(
	"																	\
		varying vec2 v_tex;												\
																		\
		void main( void )												\
		{																\
			v_tex = vec2( gl_MultiTexCoord0 );							\
																		\
			gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;		\
		}																\
																		\
	" );
	
	const std::string Cg_vert(															\
	"																					\
		void main( in  float4 v_pos:	 POSITION, in  float2 v_tex:	 TEXCOORD0,		\
				   out float4 v_out_pos: POSITION, out float2 v_out_tex: TEXCOORD0,		\
				   const uniform float4x4 model_view_proj )								\
		{																				\
			v_out_tex = v_tex;															\
			v_out_pos = mul( model_view_proj, v_pos );									\
		}																				\
	\n" );
}

n_operator_gpu_kernel::n_operator_gpu_kernel( const boost::filesystem::path& path )
	: n_operator_gpu( creator( path ) )
{
}

n_operator_gpu_kernel::~n_operator_gpu_kernel( )
{
}

#if defined ( HAVE_CG_RUNTIME )
bool n_operator_gpu_kernel::download_Cg_shader( )
{
	using namespace openpluginlib;
	
	opl::v_url* Cg_url = pointer<opl::v_url>( L"Cg_url" );
	if( Cg_url )
	{
		shader_string Cg_frag = shader_manager_instance( ).open_shader_file( Cg_url->get( ) );
		if( !is_Cg_valid( ) && shader_manager_instance( ).create_Cg_shader( Cg_vert, Cg_frag, sobj_.Cg_shader ) )
		{	
			sobj_.model_view_proj = opl::cgGetNamedParameter( sobj_.Cg_shader.vert( ), "model_view_proj" );
			sobj_.Cg_InColor0	  = opl::cgGetNamedParameter( sobj_.Cg_shader.frag( ), "inColor0" );
			
			if( sobj_.model_view_proj && sobj_.Cg_InColor0 )
				sobj_.Cg_valid = true;
		}
			
		sobj_.Cg_tried = true;
	}
	
	return is_Cg_valid( );
}

void n_operator_gpu_kernel::enable_Cg_shader( )
{
	using namespace fx;

	fx::v_node_ptr* in = pointer<fx::v_node_ptr>( L"inColor0" );

	opl::cgGLEnableProfile( sobj_.Cg_shader.v_profile( ) );
	opl::cgGLEnableProfile( sobj_.Cg_shader.f_profile( ) );

	opl::cgGLBindProgram( sobj_.Cg_shader.vert( ) );
	opl::cgGLBindProgram( sobj_.Cg_shader.frag( ) );
		
	opl::cgGLSetTextureParameter( sobj_.Cg_InColor0, in->get( )->get_GL_texture( ) );
	opl::cgGLEnableTextureParameter( sobj_.Cg_InColor0 );
		
	opl::cgGLSetStateMatrixParameter( sobj_.model_view_proj, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY );
}

void n_operator_gpu_kernel::disable_Cg_shader( )
{
	//opl::cgGLDisableTextureParameter( sobj_.Cg_InColor0 );
	opl::cgGLDisableProfile( sobj_.Cg_shader.v_profile( ) );
	opl::cgGLDisableProfile( sobj_.Cg_shader.f_profile( ) );
}
#endif

bool n_operator_gpu_kernel::download_GLSL_shader( )
{
	using namespace openpluginlib;

	opl::v_url* GLSL_url = pointer<opl::v_url>( L"GLSL_url" );
	if( GLSL_url )
	{
		shader_string GLSL_frag = shader_manager_instance( ).open_shader_file( GLSL_url->get( ) );
		if( !is_GLSL_valid( ) && shader_manager_instance( ).create_GLSL_shader( GLSL_vert, GLSL_frag, sobj_.GLSL_shader ) )
		{
			sobj_.GLSL_InColor0 = glGetUniformLocationARB( sobj_.GLSL_shader.program( ), "inColor0" );
			sobj_.GLSL_valid = true;
		}
			
		sobj_.GLSL_tried = true;
	}
	
	return is_GLSL_valid( );
}

void n_operator_gpu_kernel::enable_GLSL_shader( )
{
	glUseProgramObjectARB( sobj_.GLSL_shader.program( ) );
	glUniform1iARB( sobj_.GLSL_InColor0, 0 );
}

void n_operator_gpu_kernel::disable_GLSL_shader( )
{
	glUseProgramObjectARB( 0 );
}

bool n_operator_gpu_kernel::download_shader( )
{/*
	opl::v_url* url = pointer<opl::v_url>( L"url" );
	if( url && 
		( ( ( obj_ = opl::shader_manager_instance( ).create_shader_object( "cg:passthrough.cg", url->value( ).native_file_string( ) ) ) != 0 ) ||
		( ( ( obj_ = opl::shader_manager_instance( ).create_shader_object( "glsl:passthrough.vert", url->value( ).native_file_string( ) ) ) != 0 ) ) ) )
		return true;
*/
	return false;
}

bool n_operator_gpu_kernel::valid( ) const
{
//	if( obj_ )
//		return obj_->is_valid( );

	return false;
}

bool n_operator_gpu_kernel::init( )
{
	using namespace fx;

	fx::v_node_ptr* in = pointer<fx::v_node_ptr>( L"inColor0" );
	if( !set_size( in->get( )->width( ), in->get( )->height( ) ) )
		return false;

#if defined ( HAVE_CG_RUNTIME )
	if( !sobj_.Cg_tried )   download_Cg_shader( );
#endif

	if( !sobj_.GLSL_tried ) download_GLSL_shader( );

	return true;
}

bool n_operator_gpu_kernel::start( )
{
	glDisable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	glDisable( GL_CULL_FACE );

	rb( ).begin_render_to( );

	if( get_GL_target( ) == GL_TEXTURE_2D )
	{
		if( sobj_.GLSL_valid )
			enable_GLSL_shader( );
	}

#if defined ( HAVE_CG_RUNTIME )
	else
	{
		if( sobj_.Cg_valid )
			enable_Cg_shader( );
	}
#endif
	
	return true;
}

bool n_operator_gpu_kernel::finish( )
{
	if( get_GL_target( ) == GL_TEXTURE_2D )
	{
		if( sobj_.GLSL_valid )
			disable_GLSL_shader( );
	}

#if defined ( HAVE_CG_RUNTIME )
	else
	{
		if( sobj_.Cg_valid )
			disable_Cg_shader( );
	}
#endif

	rb( ).end_render_to( );

	return true;
}

bool n_operator_gpu_kernel::update_image( )
{
	using namespace fx;

	fx::v_node_ptr* in = pointer<fx::v_node_ptr>( L"inColor0" );
	if( !in || !in->get( ) ) return false;

	if( !in->get( )->update_image( ) || !init( ) ) return false;
	
	int rb_w		= width( );
	int rb_h		= height( );
	GLfloat tex_w	= max_s( );
	GLfloat tex_h	= max_t( );

	start( );

	glActiveTexture( GL_TEXTURE0 );
	in->get( )->bind_as_GL_texture( );
		
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0f, 0.0f );		glVertex2i( 0, 0 );
		glTexCoord2f( tex_w, 0.0f );	glVertex2i( rb_w, 0 );
		glTexCoord2f( tex_w, tex_h );	glVertex2i( rb_w, rb_h );
		glTexCoord2f( 0.0f, tex_h );	glVertex2i( 0, rb_h );
	glEnd( );

	in->get( )->unbind_as_GL_texture( );

	finish( );
	
	return true;
}

} } }
