
// fx - A compositing library representation.

// Copyright (C) 2005-2006 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#ifdef HAVE_CONFIG_H
#include <openlibraries_global_config.hpp>
#endif

#include <openpluginlib/pl/openpluginlib.hpp>
#include <openpluginlib/pl/utf8_utils.hpp>

#if defined ( HAVE_OPENIMAGELIB )
#include <openimagelib/il/openimagelib_plugin.hpp>
#endif

#include <openeffectslib/plugins/openimagelib_source/still_image_source_operator.hpp>

namespace fs  = boost::filesystem;
namespace opl = olib::openpluginlib;

namespace olib { namespace openeffectslib { namespace plugins {

namespace
{
	still_image_source_operator::container field_creator( )
	{
		typedef still_image_source_operator::container	container;
		typedef still_image_source_operator::value_type value_type;
		
		container m;

		m.insert( value_type( L"url",		opl::v_url( ) ) );
		m.insert( value_type( L"OutColor0",	fx::v_node_ptr( ) ) );

		return m;
	}
	
	still_image_source_operator::param_container param_creator( )
	{
		typedef still_image_source_operator::param_container	param_container;
		typedef still_image_source_operator::param_value_type	param_value_type;

		param_container m;
		
		m.insert( param_value_type( L"url",			opl::v_url_typename ) );
		m.insert( param_value_type( L"OutColor0",	fx::v_out_color_typename ) );
		
		return m;
	}
}

namespace
{
	struct query_traits
	{
		query_traits( const opl::wstring& filename )
			: filename_( filename )
		{ }
		
		opl::wstring libname( ) const
		{ return L"openimagelib"; }
		
		opl::wstring to_match( ) const
		{ return filename_; }

		opl::wstring type( ) const
		{ return L"input"; }
		
		int merit( ) const
		{ return 0; }
		
		opl::wstring filename_;
	};

}

still_image_source_operator::still_image_source_operator( )
	: n_operator_gpu( field_creator( ), param_creator( ) )
	, id_( 0 )
	, target_( GL_TEXTURE_2D )
{ }

bool still_image_source_operator::update_image( )
{
	opl::v_url* url = pointer<opl::v_url>( L"url" );
	if( !id_ )
		download_image( url->value( ) );

	if( !rb( ).initialised( ) ) return false;

	rb( ).begin_render_to( );

	glEnable( target_ );
	glBindTexture( target_, id_ );
	
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0f, 0.0f );
		glVertex2i( 0, 0 );
		glTexCoord2f( max_s( ), 0.0f );
		glVertex2i( width( ), 0 );
		glTexCoord2f( max_s( ), max_t( ) );
		glVertex2i( width( ), height( ) );
		glTexCoord2f( 0.0f, max_t( ) );
		glVertex2i( 0, height( ) );
	glEnd( );
	
	rb( ).end_render_to( );
	
	return true;
}

bool still_image_source_operator::download_image( const fs::path& path )
{
#if defined ( HAVE_OPENIMAGELIB )
	namespace il  = olib::openimagelib::il;
	namespace opl = olib::openpluginlib;

	typedef il::image_type::const_pointer	const_pointer;
	typedef il::image_type::size_type		size_type;

	typedef opl::discovery<query_traits> oil_discovery;

	oil_discovery plugins( query_traits( opl::to_wstring( path.native_file_string( ).c_str( ) ) ) );
	if( plugins.empty( ) ) return false;
			
	oil_discovery::const_iterator i = plugins.begin( );
	boost::shared_ptr<il::openimagelib_plugin> plug = boost::shared_static_cast<il::openimagelib_plugin>( i->create_plugin( "" ) );
	if( !plug ) return false;

	il::image_type_ptr im = plug->load( path );
	if( !im ) return false;

	im = il::conform( im, il::cropped | il::flipped | il::flopped );

	const_pointer texels	= im->data( );
	int width				= im->width( );
	int height				= im->height( );

	float dummy_tex_w, dummy_tex_h;
	if( !opl::texture_target( width, height, target_, dummy_tex_w, dummy_tex_h ) ) return false;
		
	if( id_ == 0 )
	{
		glGenTextures( 1, &id_ );
		rb( ).init( "", width, height );
	}
		
	glBindTexture( target_, id_ );
	glTexParameteri( target_, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( target_, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( target_, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( target_, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	
	typedef std::pair<GLenum, GLenum> internal_and_base_format;
	internal_and_base_format pf = opl::pf_to_gl_format( im->pf( ) );
		
	glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
		
	for( size_type i = 0; i < im->count( ); ++i )
	{
		size_type mipsize = il::calculate_mipmap_size( im, i );
	
		if( opl::is_compressed_format( im->pf( ) ) )
		{
			glCompressedTexImage2DARB( target_, static_cast<GLint>( i ), pf.second, 
				static_cast<GLsizei>( width ), static_cast<GLsizei>( height ), 0, static_cast<GLsizei>( mipsize ), texels );
		}
		else
		{
			glTexImage2D( target_, static_cast<GLint>( i ), pf.first, 
				static_cast<GLsizei>( width ), static_cast<GLsizei>( height ), 0, pf.second, GL_UNSIGNED_BYTE, texels );
		}

		texels += mipsize;
		
		width  /= 2;
		height /= 2;
	}
		
	glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
		
	return true;
#endif
}

} } }

