// pool - A pooled memory allocator

// Copyright (C) 2005 Visual Media FX Ltd.
// Released under the LGPL.
// For more information, see http://www.openlibraries.org.

#include "pool.hpp"
using namespace std;
using namespace olib::openpluginlib;

// Declaration of mutex
OPENPLUGINLIB_DECLSPEC boost::mutex pool::mutex_;

// Constructor creates 31 pools sized at 2^i
pool::pool( )
{
	for( int i = 1; i < 32; i ++ )
	{
		pools_[ i ] = new boost::pool<>( ( 1 << i ) + sizeof( int ) );
		pools_[ i ]->set_next_size( 1 );
	}
}

// Destroys pools
pool::~pool( )
{
	for( int i = 1; i < 32; i ++ )
		delete pools_[ i ];
}

// Singleton accessor - ensures pools are destroyed on exit
pool *pool::get_instance( )
{
	static pool *instance = NULL;
	boost::mutex::scoped_lock scoped_lock( mutex_ );
	if ( instance == NULL )
	{
		instance = new pool( );
		atexit( destroy );
	}
	return instance;
}

// Singleton destructor
void pool::destroy( ) 
{ 
	delete get_instance( ); 
}

// Allocate memory >= size
unsigned char *pool::malloc( int size )
{
	unsigned char *result = 0;

	if ( size > 0 )
	{
		pool *self = get_instance( );
		int index = 1;
		while ( ( 1 << index ) < size ) index ++;
		boost::mutex::scoped_lock scoped_lock( mutex_ );
		result = static_cast< unsigned char * >( self->pools_[ index ]->malloc( ) );
		if ( result != 0 )
		{
			*( int * )( result ) = index;
			result += sizeof( int );
		}
	}

	return result;
}

// Reallocate a chunk to given size
unsigned char *pool::realloc( unsigned char *chunk, int size )
{
	unsigned char *result = chunk;

	if ( chunk != 0 )
	{
		int old_index = *( int * )( chunk - sizeof( int ) );
		int index = 1;
		while ( ( 1 << index ) < size ) index ++;
		if ( old_index < index )
		{
			result = pool::malloc( size );
			memcpy( result, chunk, int( 1 << old_index ) );
			pool::free( chunk );
		}
	}
	else
	{
		result = pool::malloc( size );
	}

	return result;
}

// Free memory 
void pool::free( unsigned char *chunk )
{
	if ( chunk != 0 )
	{
		pool *self = get_instance( );
		int index = *( int * )( chunk - sizeof( int ) );
		boost::mutex::scoped_lock scoped_lock( mutex_ );
		self->pools_[ index ]->free( chunk - sizeof( int ) );
	}
}

